@@ -39,7 +39,7 @@ typedef struct crop_s
 } crop_t;
 
 void auto_crop (pic_t p, crop_t *c);
-int auto_split (pic_t p, crop_t *c);
+int auto_split (pic_t p, crop_t *c, int ugly);
 
 #endif
 
