@@ -15,6 +15,22 @@
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *----------------------------------------------------------------------------
+ * Version 2.01
+ *   - Restrict number of composition objects within an epoch to 64
+ *
+ * Version 2.00
+ *   - Options --seek and --count were added, to allow for partial processing
+ *     of input video
+ *   - For debug purposes, a stand-alone SUP file parser was added in debug/
+ *   - Experimental SUP output support
+ *
+ * Version 1.13
+ *   - By default, images are no longer split into to composition objects,
+ *     if the result would look ugly. For example, descenders should no
+ *     longer get put into their own image. Reverting back to the old
+ *     behaviour is possible using the --ugly option.
+ *   - Ensure minimum size of 8x8 for composition objects
+ *
  * Version 1.12
  *   - Fix error in example
  *   - Fix double free, which could result in broken images when image
@@ -90,19 +106,24 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <limits.h>
 #include <time.h>
 #include <png.h>
 #include <getopt.h>
+#include <assert.h>
 #include "auto_split.h"
 #include "palletize.h"
+#include "sup.h"
 
 /* AVIS input code taken from muxers.c from the x264 project (GPLv2 or later).
  * Authors: Laurent Aimar <fenrir@via.ecp.fr>
  *          Loren Merritt <lorenm@u.washington.edu>
  * Slightly modified, so if you find breakage, it's probably mine. :)
  */
+#ifndef LINUX
 #include <windows.h>
 #include <vfw.h>
+#endif
 
 static int64_t gcd( int64_t a, int64_t b )
 {
@@ -117,7 +138,9 @@ static int64_t gcd( int64_t a, int64_t b )
 }
 
 typedef struct {
+#ifndef LINUX
     PAVISTREAM p_avi;
+#endif
     int width, height;
 } avis_input_t;
 
@@ -130,6 +153,7 @@ typedef struct {
 
 int open_file_avis( char *psz_filename, avis_input_t **p_handle, stream_info_t *p_param )
 {
+#if !defined(LINUX)
     avis_input_t *h = malloc(sizeof(avis_input_t));
     AVISTREAMINFO info;
     int i;
@@ -177,10 +201,19 @@ int open_file_avis( char *psz_filename, avis_input_t **p_handle, stream_info_t *
         (int)info.dwLength );
 
     return 0;
+#else
+    p_param->i_width = 1920;
+    p_param->i_height = 1080;
+    p_param->i_fps_den = 30000;
+    p_param->i_fps_num = 1001;
+    *p_handle = fopen(psz_filename, "r");
+		return 0;
+#endif
 }
 
 int get_frame_total_avis( avis_input_t *handle )
 {
+#if !defined(LINUX)
     avis_input_t *h = handle;
     AVISTREAMINFO info;
 
@@ -188,25 +221,38 @@ int get_frame_total_avis( avis_input_t *handle )
         return -1;
 
     return info.dwLength;
+#else
+    return 6016;
+#endif
 }
 
 int read_frame_avis( char *p_pic, avis_input_t *handle, int i_frame )
 {
+#if !defined(LINUX)
     avis_input_t *h = handle;
 
     if( AVIStreamRead(h->p_avi, i_frame, 1, p_pic, h->width * h->height * 4, NULL, NULL ) )
         return -1;
 
     return 0;
+#else
+    fread(p_pic, 4, 1920*1080, handle);
+    return 0;
+#endif
 }
 
 int close_file_avis( avis_input_t *handle )
 {
+#if !defined(LINUX)
     avis_input_t *h = handle;
     AVIStreamRelease(h->p_avi);
     AVIFileExit();
     free(h);
     return 0;
+#else
+    fclose(handle);
+    return 0;
+#endif
 }
 
 /* AVIS code ends here */
@@ -517,6 +563,9 @@ int detect_sse2 ()
 	unsigned int func = 0x00000001;
 	unsigned int eax, ebx, ecx, edx;
 
+	/* FIXME: SSE2 broken with new GCC. */
+	/* return 0; */
+
 	if (detection != -1)
 		return detection;
 
@@ -606,6 +655,9 @@ void print_usage ()
 		"Usage: avs2bdnxml [options] -o output input\n\n"
 		"Input has to be an AviSynth script with RGBA as output colorspace\n\n"
 		"  -o, --output <string>        Output file in BDN XML format\n"
+		"                               For SUP/PGS output, use a .sup extension\n"
+		"  -j, --seek <integer>         Start processing at this frame, first is 0\n"
+		"  -c, --count <integer>        Number of input frames to process\n"
 		"  -t, --trackname <string>     Name of track, like: Undefined\n"
 		"  -l, --language <string>      Language code, like: und\n"
 		"  -v, --video-format <string>  Either of: 480i, 480p,  576i,\n"
@@ -618,15 +670,33 @@ void print_usage ()
 		"  -m, --min-split <integer>    Minimum length of line segment after split.\n"
 		"  -a, --autocrop <integer>     Automatically crop output. [on=1, off=0]\n"
 		"  -p, --palette <integer>      Output 8bit palette PNG. [on=1, off=0]\n"
+		"  -u, --ugly <integer>         Allow splitting images in ugly ways.\n"
+		"                               Might improve buffer problems, but is ugly.\n"
+		"                               [on=1, off=0]\n"
 		"  -b, --buffer-opt <integer>   Optimize PG buffer size by image\n"
 		"                               splitting. [on=1, off=0]\n\n"
 		"Example:\n"
 		"  avs2bdnxml -t Undefined -l und -v 1080p -f 23.976 -a1 -p1 -b0 -m3 \\\n"
-		"    -o output.xml input.avs\n"
+		"    -u0 -o output.xml input.avs\n"
 		"  (Input and output are required settings. The rest are set to default.)\n"
 		);
 }
 
+extern char *rindex(const char *s, int c);
+int is_extension(char *filename, char *check_ext)
+{
+	char *ext = rindex(filename, '.');
+
+	if (ext == NULL)
+		return 0;
+
+	ext++;
+	if (!strcasecmp(ext, check_ext))
+		return 1;
+
+	return 0;
+}
+
 struct event_list_node_s
 {
 	int image_number;
@@ -684,27 +754,50 @@ void add_event_xml (event_list_head_t *events, int split_at, int min_split, int
 	}
 }
 
+void write_sup_wrapper (sup_writer_t *sw, uint8_t *im, int num_crop, crop_t *crops, uint32_t *pal, int start, int end, int split_at, int min_split)
+{
+	int image = start;
+	int d = end - start;
+
+	if (!split_at)
+		write_sup(sw, im, num_crop, crops, pal, start, end);
+	else
+	{
+		while (d >= split_at + min_split)
+		{
+			d -= split_at;
+			write_sup(sw, im, num_crop, crops, pal, start, start + split_at);
+			start += split_at;
+		}
+		if (d)
+			write_sup(sw, im, num_crop, crops, pal, start, start + d);
+	}
+}
+
+
 struct framerate_entry_s
 {
 	char *name;
 	char *out_name;
 	int rate;
 	int drop;
+	int fps_num;
+	int fps_den;
 };
 
 /* Most of the time seems to be spent in AviSynth (about 4/5). */
 int main (int argc, char *argv[])
 {
-	struct framerate_entry_s framerates[] = { {"23.976", "23.976", 24, 0}
+	struct framerate_entry_s framerates[] = { {"23.976", "23.976", 24, 0, 24000, 1001}
 	                                        /*, {"23.976d", "23.976", 24000/1001.0, 1}*/
-	                                        , {"24", "24", 24, 0}
-	                                        , {"25", "25", 25, 0}
-	                                        , {"29.97", "29.97", 30, 0}
+	                                        , {"24", "24", 24, 0, 24, 1}
+	                                        , {"25", "25", 25, 0, 25, 1}
+	                                        , {"29.97", "29.97", 30, 0, 30000, 1001}
 	                                        /*, {"29.97d", "29.97", 30000/1001.0, 1}*/
-	                                        , {"50", "50", 50, 0}
-	                                        , {"59.94", "59.94", 60, 0}
+	                                        , {"50", "50", 50, 0, 50, 1}
+	                                        , {"59.94", "59.94", 60, 0, 60000, 1001}
 	                                        /*, {"59.94d", "59.94", 60000/1001.0, 1}*/
-	                                        , {NULL, NULL, 0, 0}
+	                                        , {NULL, NULL, 0, 0, 0, 0}
 	                                        };
 	char *avs_filename = NULL;
 	char *track_name = "Undefined";
@@ -719,6 +812,9 @@ int main (int argc, char *argv[])
 	char *minimum_split = "3";
 	char *palletize_png = "1";
 	char *auto_crop_image = "1";
+	char *ugly_option = "0";
+	char *seek_string = "0";
+	char *count_string = "2147483647";
 	char *in_img = NULL, *old_img = NULL, *tmp = NULL, *out_buf = NULL;
 	char *intc_buf = NULL, *outtc_buf = NULL;
 	char *drop_frame = NULL;
@@ -732,6 +828,8 @@ int main (int argc, char *argv[])
 	int autocrop = 0;
 	int xo, yo;
 	int fps = 25;
+	int count_frames = INT_MAX, last_frame;
+	int init_frame = 0;
 	int frames;
 	int first_frame = -1, start_frame = -1, end_frame = -1;
 	int num_of_events = 0;
@@ -741,9 +839,14 @@ int main (int argc, char *argv[])
 	int checked_empty;
 	int auto_cut = 0;
 	int pal_png = 1;
+	int ugly = 0;
 	int progress_step = 1000;
 	int buffer_opt;
 	int bench_start = time(NULL);
+	int fps_num = 25, fps_den = 1;
+	int sup_output = 0;
+	int xml_output = 0;
+	sup_writer_t *sw;
 	avis_input_t *avis_hnd;
 	stream_info_t *s_info = malloc(sizeof(stream_info_t));
 	event_list_head_t *events = malloc(sizeof(event_list_head_t));
@@ -760,6 +863,8 @@ int main (int argc, char *argv[])
 	{
 		static struct option long_options[] =
 			{ {"output",       required_argument, 0, 'o'}
+			, {"seek",         required_argument, 0, 'j'}
+			, {"count",        required_argument, 0, 'c'}
 			, {"trackname",    required_argument, 0, 't'}
 			, {"language",     required_argument, 0, 'l'}
 			, {"video-format", required_argument, 0, 'v'}
@@ -771,11 +876,12 @@ int main (int argc, char *argv[])
 			, {"autocrop",     required_argument, 0, 'a'}
 			, {"palette",      required_argument, 0, 'p'}
 			, {"buffer-opt",   required_argument, 0, 'b'}
+			, {"ugly",         required_argument, 0, 'u'}
 			, {0, 0, 0, 0}
 			};
 			int option_index = 0;
 
-			c = getopt_long(argc, argv, "o:t:l:v:f:x:y:b:s:m:p:a:", long_options, &option_index);
+			c = getopt_long(argc, argv, "o:j:c:t:l:v:f:x:y:b:s:m:p:a:u:", long_options, &option_index);
 			if (c == -1)
 				break;
 			switch (c)
@@ -783,6 +889,12 @@ int main (int argc, char *argv[])
 				case 'o':
 					out_filename = optarg;
 					break;
+				case 'j':
+					seek_string = optarg;
+					break;
+				case 'c':
+					count_string = optarg;
+					break;
 				case 't':
 					track_name = optarg;
 					break;
@@ -816,6 +928,9 @@ int main (int argc, char *argv[])
 				case 'm':
 					minimum_split = optarg;
 					break;
+				case 'u':
+					ugly_option = optarg;
+					break;
 				default:
 					print_usage();
 					return 0;
@@ -842,12 +957,33 @@ int main (int argc, char *argv[])
 		return 0;
 	}
 
+	/* Get target output format */
+	if (is_extension(out_filename, "xml"))
+	{
+		xml_output = 1;
+		sup_output = 0;
+	}
+	else if (is_extension(out_filename, "sup") || is_extension(out_filename, "pgs"))
+	{
+		sup_output = 1;
+		xml_output = 0;
+		pal_png = 1;
+	}
+	else
+	{
+		fprintf(stderr, "Output file extension must be \".xml\", \".sup\" or \".pgs\".\n");
+		return 1;
+	}
+
 	/* Set X and Y offsets, and split value */
 	xo = atoi(x_offset);
 	yo = atoi(y_offset);
 	pal_png = atoi(palletize_png);
 	autocrop = atoi(auto_crop_image);
 	split_at = atoi(split_after);
+	ugly = atoi(ugly_option);
+	init_frame = atoi(seek_string);
+	count_frames = atoi(count_string);
 	min_split = atoi(minimum_split);
 	if (!min_split)
 		min_split = 1;
@@ -863,6 +999,8 @@ int main (int argc, char *argv[])
 			frame_rate = framerates[i].out_name;
 			fps = framerates[i].rate;
 			drop_frame = framerates[i].drop ? "true" : "false";
+			fps_num = framerates[i].fps_num;
+			fps_den = framerates[i].fps_den;
 			have_fps = 1;
 		}
 		i++;
@@ -886,6 +1024,13 @@ int main (int argc, char *argv[])
 	old_img = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char)); /* see above */
 	out_buf = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char));
 
+	/* Check minimum size */
+	if (s_info->i_width < 8 || s_info->i_height < 8)
+	{
+		fprintf(stderr, "Error: Video dimensions below 8x8 (%dx%d).\n", s_info->i_width, s_info->i_height);
+		return 1;
+	}
+
 	/* Align buffers */
 	in_img  = in_img + (16 - ((int)in_img % 16));
 	old_img = old_img + (16 - ((int)old_img % 16));
@@ -904,26 +1049,34 @@ int main (int argc, char *argv[])
 	crops[0].h = pic.h;
 
 	/* Get frame number */
-	frames  = get_frame_total_avis(avis_hnd);
+	frames = get_frame_total_avis(avis_hnd);
+	if (count_frames + init_frame > frames)
+	{
+		count_frames = frames - init_frame;
+	}
+	last_frame = count_frames + init_frame;
 
 	/* No frames mean nothing to do */
-	if (frames < 1)
+	if (count_frames < 1)
 	{
 		fprintf(stderr, "No frames found.\n");
 		return 0;
 	}
 
 	/* Set progress step */
-	if (frames < 1000)
+	if (count_frames < 1000)
 	{
-		if (frames > 200)
+		if (count_frames > 200)
 			progress_step = 50;
-		else if (frames > 50)
+		else if (count_frames > 50)
 			progress_step = 10;
 		else
 			progress_step = 1;
 	}
 
+	/* Open SUP writer, if applicable */
+	sw = new_sup_writer(out_filename, pic.w, pic.h, fps_num, fps_den);
+
 	/* Initialize events */
 	events->start = NULL;
 	events->end = NULL;
@@ -932,7 +1085,7 @@ int main (int argc, char *argv[])
 		asm ("emms\n");
 
 	/* Process frames */
-	for (i = 0; i < frames; i++)
+	for (i = init_frame; i < last_frame; i++)
 	{
 		if (read_frame_avis(in_img, avis_hnd, i))
 		{
@@ -942,9 +1095,9 @@ int main (int argc, char *argv[])
 		checked_empty = 0;
 
 		/* Progress indicator */
-		if (i % (frames / progress_step) == 0)
+		if (i % (count_frames / progress_step) == 0)
 		{
-			fprintf(stderr, "\rProgress: %d/%d - Lines: %d", i, frames, num_of_events);
+			fprintf(stderr, "\rProgress: %d/%d - Lines: %d", i - init_frame, count_frames, num_of_events);
 		}
 
 		/* If we are outside any lines, check for empty frames first */
@@ -957,7 +1110,7 @@ int main (int argc, char *argv[])
 		}
 
 		/* Check for duplicate, unless first frame */
-		if (i && have_line && is_identical(s_info, in_img, old_img))
+		if ((i != init_frame) && have_line && is_identical(s_info, in_img, old_img))
 			continue;
 		/* Mark frames that were not used as new image in comparison to have transparent pixels zeroed */
 		else if (!(i && have_line))
@@ -966,7 +1119,15 @@ int main (int argc, char *argv[])
 		/* Not a dup, write end-of-line, if we had a line before */
 		if (have_line)
 		{
-			add_event_xml(events, split_at, min_split, start_frame, i, n_crop, crops);
+			if (sup_output)
+			{
+				assert(pal != NULL);
+				write_sup_wrapper(sw, out_buf, n_crop, crops, pal, start_frame, i, split_at, min_split);
+				free(pal);
+				pal = NULL;
+			}
+			else if (xml_output)
+				add_event_xml(events, split_at, min_split, start_frame, i, n_crop, crops);
 			end_frame = i;
 			have_line = 0;
 		}
@@ -985,7 +1146,7 @@ int main (int argc, char *argv[])
 		start_frame = i;
 		swap_rb(s_info, in_img, out_buf);
 		if (buffer_opt)
-			n_crop = auto_split(pic, crops);
+			n_crop = auto_split(pic, crops, ugly);
 		else if (autocrop)
 		{
 			crops[0].x = 0;
@@ -996,9 +1157,10 @@ int main (int argc, char *argv[])
 		}
 		if (pal_png)
 			pal = palletize(out_buf, s_info->i_width, s_info->i_height);
-		for (j = 0; j < n_crop; j++)
-			write_png(start_frame, out_buf, s_info->i_width, s_info->i_height, j, pal, crops[j]);
-		if (pal_png)
+		if (xml_output)
+			for (j = 0; j < n_crop; j++)
+				write_png(start_frame, out_buf, s_info->i_width, s_info->i_height, j, pal, crops[j]);
+		if (pal_png && !sup_output)
 		{
 			free(pal);
 			pal = NULL;
@@ -1013,85 +1175,103 @@ int main (int argc, char *argv[])
 		old_img = tmp;
 	}
 
-	fprintf(stderr, "\rProgress: %d/%d - Lines: %d - Done\n", i, frames, num_of_events);
+	fprintf(stderr, "\rProgress: %d/%d - Lines: %d - Done\n", i - init_frame, count_frames, num_of_events);
 
 	/* Add last event, if available */
 	if (have_line)
 	{
-		add_event_xml(events, split_at, min_split, start_frame, i - 1, n_crop, crops);
+		if (sup_output)
+		{
+			assert(pal != NULL);
+			write_sup_wrapper(sw, out_buf, n_crop, crops, pal, start_frame, i - 1, split_at, min_split);
+			free(pal);
+			pal = NULL;
+		}
+		else if (xml_output)
+			add_event_xml(events, split_at, min_split, start_frame, i - 1, n_crop, crops);
 		auto_cut = 1;
 		end_frame = i - 1;
 	}
 
-	/* Check if we actually have any events */
-	if (first_frame == -1)
+	if (sup_output)
 	{
-		fprintf(stderr, "No events detected. Refusing to write XML file.\n");
-		return 1;
+		close_sup_writer(sw);
 	}
 
-	if (detect_sse2())
-		asm ("emms\n");
-
-	/* Initialize timecode buffers */
-	intc_buf = calloc(12, 1);
-	outtc_buf = calloc(12, 1);
-
-	/* Creating XML file */
-	if ((fh = fopen(out_filename, "w")) == 0)
+	if (xml_output)
 	{
-		perror("Error opening output file");
-		return 1;
-	}
-
-	/* Write XML header */
-	mk_timecode(first_frame, fps, intc_buf);
-	mk_timecode(end_frame, fps, outtc_buf);
-	fprintf(fh, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
-		"<BDN Version=\"0.93\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
-		"xsi:noNamespaceSchemaLocation=\"BD-03-006-0093b BDN File Format.xsd\">\n"
-		"<Description>\n"
-		"<Name Title=\"%s\" Content=\"\"/>\n"
-		"<Language Code=\"%s\"/>\n"
-		"<Format VideoFormat=\"%s\" FrameRate=\"%s\" DropFrame=\"%s\"/>\n"
-		"<Events LastEventOutTC=\"%s\" FirstEventInTC=\"%s\"\n", track_name, language, video_format, frame_rate, drop_frame, outtc_buf, intc_buf);
-
-	mk_timecode(0, fps, intc_buf);
-	mk_timecode(frames, fps, outtc_buf);
-	fprintf(fh, "ContentInTC=\"%s\" ContentOutTC=\"%s\" NumberofEvents=\"%d\" Type=\"Graphic\"/>\n"
-		"</Description>\n"
-		"<Events>\n", intc_buf, outtc_buf, num_of_events);
-
-	/* Write XML events */
-	if (events->start != NULL)
-	{
-		node = events->start;
-		do
+		/* Check if we actually have any events */
+		if (first_frame == -1)
 		{
-			mk_timecode(node->start_frame, fps, intc_buf);
-			mk_timecode(node->end_frame, fps, outtc_buf);
-			
-			if (auto_cut && node->end_frame == frames - 1)
-			{
-				mk_timecode(node->end_frame + 1, fps, outtc_buf);
-			}
-			
-			fprintf(fh, "<Event Forced=\"False\" InTC=\"%s\" OutTC=\"%s\">\n", intc_buf, outtc_buf);
-			for (i = 0; i < node->graphics; i++)
-			{
-				fprintf(fh, "<Graphic Width=\"%d\" Height=\"%d\" X=\"%d\" Y=\"%d\">%08d_%d.png</Graphic>\n", node->c[i].w, node->c[i].h, xo + node->c[i].x, yo + node->c[i].y, node->image_number, i);
-			}
-			fprintf(fh, "</Event>\n");
-			node = node->next;
+			fprintf(stderr, "No events detected. Refusing to write XML file.\n");
+			return 1;
 		}
-		while (node != NULL);
-	}
 
-	/* Write XML footer */
-	fprintf(fh, "</Events>\n</BDN>\n");
+		if (detect_sse2())
+			asm ("emms\n");
+
+		/* Initialize timecode buffers */
+		intc_buf = calloc(12, 1);
+		outtc_buf = calloc(12, 1);
+
+		/* Creating output file */
+		if ((fh = fopen(out_filename, "w")) == 0)
+		{
+			perror("Error opening output XML file");
+			return 1;
+		}
+
+		/* Write XML header */
+		mk_timecode(first_frame, fps, intc_buf);
+		mk_timecode(end_frame, fps, outtc_buf);
+		fprintf(fh, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+			"<BDN Version=\"0.93\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n"
+			"xsi:noNamespaceSchemaLocation=\"BD-03-006-0093b BDN File Format.xsd\">\n"
+			"<Description>\n"
+			"<Name Title=\"%s\" Content=\"\"/>\n"
+			"<Language Code=\"%s\"/>\n"
+			"<Format VideoFormat=\"%s\" FrameRate=\"%s\" DropFrame=\"%s\"/>\n"
+			"<Events LastEventOutTC=\"%s\" FirstEventInTC=\"%s\"\n", track_name, language, video_format, frame_rate, drop_frame, outtc_buf, intc_buf);
+
+		mk_timecode(0, fps, intc_buf);
+		mk_timecode(frames, fps, outtc_buf);
+		fprintf(fh, "ContentInTC=\"%s\" ContentOutTC=\"%s\" NumberofEvents=\"%d\" Type=\"Graphic\"/>\n"
+			"</Description>\n"
+			"<Events>\n", intc_buf, outtc_buf, num_of_events);
+
+		/* Write XML events */
+		if (events->start != NULL)
+		{
+			node = events->start;
+			do
+			{
+				mk_timecode(node->start_frame, fps, intc_buf);
+				mk_timecode(node->end_frame, fps, outtc_buf);
+				
+				if (auto_cut && node->end_frame == frames - 1)
+				{
+					mk_timecode(node->end_frame + 1, fps, outtc_buf);
+				}
+				
+				fprintf(fh, "<Event Forced=\"False\" InTC=\"%s\" OutTC=\"%s\">\n", intc_buf, outtc_buf);
+				for (i = 0; i < node->graphics; i++)
+				{
+					fprintf(fh, "<Graphic Width=\"%d\" Height=\"%d\" X=\"%d\" Y=\"%d\">%08d_%d.png</Graphic>\n", node->c[i].w, node->c[i].h, xo + node->c[i].x, yo + node->c[i].y, node->image_number, i);
+				}
+				fprintf(fh, "</Event>\n");
+				node = node->next;
+			}
+			while (node != NULL);
+		}
+
+		/* Write XML footer */
+		fprintf(fh, "</Events>\n</BDN>\n");
+
+		/* Close XML file */
+		fclose(fh);
+	}
 
 	/* Cleanup */
-	fclose(fh);
 	close_file_avis(avis_hnd);
 
 	/* Give runtime */
